local MAJOR = "LibUnitID-1.0"
--@alpha@
MAJOR = MAJOR.."-alpha"
--@end-alpha@
local MINOR = 9999
--@non-debug@
MINOR = 3
--@end-non-debug@
local LibUnitID,libminor = LibStub(MAJOR,MINOR)

-- set a default chat command hook in case we bail out
SlashCmdList["LIBUNITID"] = function()
	print("LibUnitID: Disembedded utilities for rev. "..MINOR.." are not compatible with rev. "..libminor..".  Please update your standalone '!LibUnitID' to use the utilities.")
end
SLASH_LIBUNITID1 = "/libunitid"
SLASH_LIBUNITID2 = "/lui"

-- bail out if a newer copy was already loaded somehow
if (libminor ~= MINOR) then
	SlashCmdList["LIBUNITID"]()
	return
end

-- check which addons already loaded, and thus probably already localized UnitGUID
local preloaded = {}
do
	local n = GetNumAddOns()
	if (n > 0) then
		for i=1,n do
			--local name,title,notes,enabled,loadable,reason,security = GetAddOnInfo(i)
			preloaded[n] = (GetAddOnInfo(i))
		end
	end
end

-- initialize vars
local logSince = GetTime()
local logAPI = 0
local logCacheHits = 0
local timeLoop
local timeAPI
local timeCacheUpdate
local timeCacheHit

-- store initial function pointers
local API_UnitGUID = UnitGUID
local lib_GetUnitGUID = LibUnitID.GetUnitGUID
local lib_UnitGUID = LibUnitID.UnitGUID

-- set the traps
local relay_API_UnitGUID = API_UnitGUID
UnitGUID = function(...) return relay_API_UnitGUID(...) end
_G["UnitGUID"] = function(...) return relay_API_UnitGUID(...) end

local relay_lib_GetUnitGUID = lib_GetUnitGUID
LibUnitID.GetUnitGUID = relay_lib_GetUnitGUID

local relay_lib_UnitGUID = lib_UnitGUID
LibUnitID.UnitGUID = relay_lib_UnitGUID

-- define some wrappers
local log_API_UnitGUID = function(...)
	logAPI = logAPI + 1
	return API_UnitGUID(...)
end

local log_lib_GetUnitGUID = function(...)
	logCacheHits = logCacheHits + 1
	return lib_GetUnitGUID(...)
end

local log_lib_UnitGUID = function(...)
	logCacheHits = logCacheHits + 1
	return lib_UnitGUID(...)
end

-- monitor/eventhandler frame
local frame = CreateFrame("Frame", nil, UIParent)

local function monitor_update(delta)
	--print("update")
end -- monitor_update()

local function monitor_OnShow()
	LibUnitID:RegisterAll(monitor_update)
end -- monitor_OnShow()

local function monitor_OnHide()
	LibUnitID:UnregisterAll(monitor_update)
end -- monitor_OnHide()

-- when savedvars load, initialize logging and monitoring
frame:SetScript("OnEvent", function(self, event, arg1)
	if (arg1 == "!LibUnitID") then
		if (not LibUnitID_DB) then
			LibUnitID_DB = {}
		end
		if (LibUnitID_DB.log) then
			relay_API_UnitGUID = log_API_UnitGUID
			relay_lib_GetUnitGUID = log_lib_GetUnitGUID
			relay_lib_UnitGUID = log_lib_UnitGUID
		end
		frame:UnregisterEvent("ADDON_LOADED")
		frame:SetScript("OnEvent", nil)
		frame:Hide()
		frame:SetScript("OnShow", monitor_OnShow)
		frame:SetScript("OnHide", monitor_OnHide)
		if (LibUnitID_DB.monitor) then
			frame:Show()
		end
	end
end)
frame:RegisterEvent("ADDON_LOADED")


--[[ TIMING TEST ]]--

local function speedtest()
	-- TODO: repeat test to fine tune results
	local GetTime = GetTime
	local n,t,func
	-- sanity check
	if (not API_UnitGUID("target")) then return false end
	-- timing loop overhead
	n = 128
	repeat
		n = n * 2
		t = GetTime()
		for i=1,n do end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeLoop = (t * 1e9) / n -- nanoseconds per loop
	-- API call
	func = API_UnitGUID
	n = 128
	repeat
		local func = func
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeAPI = ((t * 1e9) / n) - timeLoop -- nanoseconds per loop
	-- library cache update
	do
		local API_UnitGUID = API_UnitGUID
		local fakeTagID = { target=1 }
		local tagID = { target=1 }
		local delta = {}
		local fakeDeltaSize = 0
		local deltaSize = 1
		func = function(tag)
			local oldID = tagID[tag]
			if (oldID == nil) then return end
			local newID = API_UnitGUID(tag) or false
			if (newID == oldID) then return end
			fakeTagID[tag] = newID
			fakeDeltaSize = fakeDeltaSize + 1
			delta[deltaSize] = tag
			if (delta[tag]) then return end
			if (delta[tag]) then return end
		end
	end
	n = 128
	repeat
		local func = func
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeCacheUpdate = ((t * 1e9) / n) - timeLoop -- nanoseconds per loop
	-- library cache hit
	do
		local tagID = { target=API_UnitGUID("target") }
		local logCacheHits = 0
		func = function(unitid)
			return tagID[unitid]
		end
	end
	n = 128
	repeat
		local func = func
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeCacheHit = ((t * 1e9) / n) - timeLoop -- nanoseconds per loop
	-- done
	return true
end -- speedtest()


--[[ COMMAND PARSER ]]--

local function s(f) return (f<0.001 and format("%.3f",f*1e6).."us") or (f<1 and format("%.3f",f*1e3).."ms") or format("%.3f",f).."s" end
local function ms(f) return (f<1 and format("%.3f",f*1e3).."us") or (f<1e3 and format("%.3f",f).."ms") or format("%.3f",f/1e3).."s" end
local function us(f) return (f<1e3 and format("%.3f",f).."us") or (f<1e6 and format("%.3f",f/1e3).."ms") or format("%.3f",f/1e6).."s" end
local function ns(f) return (f<1e3 and format("%.3f",f).."ns") or (f<1e6 and format("%.3f",f/1e3).."us") or format("%.3f",f/1e6).."ms" end
local function pct(f) return format("%.3f",f).."%" end

SlashCmdList["LIBUNITID"] = function(args)
	-- make sure the library didn't upgrade
	local LibUnitID,libminor = LibStub(MAJOR,MINOR)
	if (libminor ~= MINOR) then
		print("LibUnitID: Disembedded utilities for rev. "..MINOR.." are not compatible with rev. "..libminor..".  Please update your standalone '!LibUnitID' to use the utilities.")
		return
	end
	-- process command
	if (strsub(args,1,3) == "log") then
		local log = strsub(args,5)
		if (log == "") then
			print("LibUnitID: Logging is currently "..((LibUnitID_DB.log and "on") or "off")..".")
		elseif (log == "on") then
			if (LibUnitID_DB.log) then
				print("LibUnitID: Logging is already on.")
			else
				print("LibUnitID: API and library UnitGUID() calls are now logged.")
				if (#preloaded > 0) then print("  "..#preloaded.." addons loaded first and might bypass the hooks; check '/libunitid list'.") end
				relay_API_UnitGUID = log_API_UnitGUID
				relay_lib_GetUnitGUID = log_lib_GetUnitGUID
				relay_lib_UnitGUID = log_lib_UnitGUID
				LibUnitID_DB.log = true
			end
		elseif (log == "off") then
			if (LibUnitID_DB.log) then
				print("LibUnitID: API and library UnitGUID() calls are no longer logged.")
				relay_API_UnitGUID = API_UnitGUID
				relay_lib_GetUnitGUID = lib_GetUnitGUID
				relay_lib_UnitGUID = lib_UnitGUID
				LibUnitID_DB.log = nil
			else
				print("LibUnitID: Logging is already off.")
			end
		end
	elseif (args == "time") then
		if (speedtest()) then
			print("Nanoseconds per operation (seconds per billion operations):")
			print("  "..ns(timeAPI).." per API UnitGUID()")
			print("  "..ns(timeCacheUpdate).." per Cache Update")
			print("  "..ns(timeCacheHit).." per Cached UnitGUID()")
		else
			print("LibUnitID: You must have a target to generate timing data; target something and try again.")
		end
    elseif (args == "report") then
		local interval = GetTime() - logSince -- seconds
		print("LibUnitID: Operations in the last "..s(interval)..":")
		local logCacheUpdates = LibUnitID:GetStatistics()
		print("  "..logAPI.." API UnitGUID()s")
		print("  "..logCacheUpdates.." Cache Updates")
		print("  "..logCacheHits.." Cached UnitGUID()s")
		if (timeAPI and timeCacheUpdate and timeCacheHit) then
			print("Nanoseconds per operation (seconds per billion operations):")
			print("  "..ns(timeAPI).." per API UnitGUID()")
			print("  "..ns(timeCacheUpdate).." per Cache Update")
			print("  "..ns(timeCacheHit).." per Cached UnitGUID()")
			print("Analysis:")
			local cost = (logCacheUpdates * timeCacheUpdate) / 1e3 -- milliseconds
			print("  "..ms(cost).." spent updating the library's cache")
			local benefit = ((logCacheHits * timeAPI) / 1e3) - ((logCacheHits * timeCacheHit) / 1e3) -- milliseconds
			print("  "..ms(benefit).." saved by hitting the cache instead of the API")
			local net = benefit - cost -- milliseconds
			local efficiency = ((net / 1e3) * 100.0) / interval
			if (net < 0) then
				print("  "..pct(-efficiency).." overall loss ("..ms(-net).." extra spent over the last "..s(interval)..")")
			else
				print("  "..pct(efficiency).." overall gain ("..ms(net).." saved over the last "..s(interval)..")")
			end
			local bestBenefit = ((logAPI * timeAPI) / 1e3) - ((logAPI * timeCacheHit) / 1e3) -- milliseconds
			print("  "..ms(bestBenefit).." more could have been saved if all native UnitGUID() calls had used the cache")
			local bestNet = net + bestBenefit -- milliseconds
			local bestEfficiency = ((bestNet / 1e3) * 100.0) / interval
			if (bestNet < 0) then
				print("  "..pct(-bestEfficiency).." best-case loss ("..ms(-bestNet).." extra still spent over the last "..s(interval)..")")
			else
				print("  "..pct(bestEfficiency).." best-case gain ("..ms(bestNet).." could have been saved over the last "..s(interval)..")")
			end
		else
			print("Efficiency analysis not available without timing data; run '/libunitid time'")
		end
	elseif (args == "reset") then
		logSince = GetTime()
		logAPI = 0
		logCacheHits = 0
		LibUnitID:ResetStatistics()
		print("LibUnitID: Statistics reset.")
	elseif (args == "list") then
		local n = #preloaded
		if (n == 0) then
			print("LibUnitID: No addons loaded before LibUnitID; all UnitGUID() calls should be trapped.")
		else
			print("LibUnitID: "..n.." addons loaded before the LibUnitID, and their UnitGUID() calls might bypass the logging hooks:")
			local s = ""
			for n,name in ipairs(preloaded) do
				s = s..", "..name
			end
			print("  "..strsub(s,3))
		end
	elseif (strsub(args,1,7) == "monitor") then
		local monitor = strsub(args,9)
		if (monitor == "") then
			print("LibUnitID: Monitoring is currently "..((LibUnitID_DB.monitor and "on") or "off")..".")
		elseif (monitor == "on") then
			if (LibUnitID_DB.monitor) then
				print("LibUnitID: Monitoring is already on.")
			else
				print("LibUnitID: All supported unitids are now monitored.")
				frame:Show()
				LibUnitID_DB.monitor = true
			end
		elseif (monitor == "off") then
			if (LibUnitID_DB.monitor) then
				print("LibUnitID: Monitoring disabled.")
				frame:Hide()
				LibUnitID_DB.monitor = nil
			else
				print("LibUnitID: Monitoring is already off.")
			end
		end
    else
		print("/libunitid < log [on | off] | time | report | reset | list | monitor [on | off] >")
    end
end

